import 'dart:collection';
import 'dart:io';

import 'package:dart_style/dart_style.dart';
import 'package:flavored_env/src/config.dart';
import 'package:flavored_env/src/models/env_var.dart';
import 'package:flavored_env/src/models/flavor.dart';
import 'package:flavored_env/src/utils.dart';
import 'package:path/path.dart';
import 'package:yaml/yaml.dart';

class FlavoredEnv {
  FlavoredEnv({
    this.description = '',
    this.envPaths = const [],
    this.classPath = const [],
    this.className = '',
    this.environments = const {},
  });

  factory FlavoredEnv.fromYamlMap(YamlMap yamlMap) {
    assert(yamlMap.runtimeType == YamlMap, 'yamlMap is not a YamlMap object');

    final YamlMap flavorsMap = yamlMap['flavors'] as YamlMap;

    final List<MapEntry<String, Flavor>> entries = flavorsMap.entries
        .map<MapEntry<String, Flavor>>(
          (e) => MapEntry<String, Flavor>(
            e.key as String,
            Flavor.fromYamlMap(e.key as String, e.value as YamlMap? ?? YamlMap()),
          ),
        )
        .toList();

    return FlavoredEnv(
      description: yamlMap[infoField] as String? ?? '',
      envPaths: getList(yamlMap[dotenvTo] as YamlList?) ?? defaultDotenvPath,
      classPath: getList(yamlMap[classTo] as YamlList?) ?? defaultClassPath,
      className: yamlMap['class_name'] as String? ?? kEnvironmentClass,
      environments: Map.fromEntries(entries),
    );
  }

  final String className;
  final List<String> classPath;
  final String description;
  final List<String> envPaths;
  final Map<String, Flavor> environments;

  static List<String>? getList(YamlList? yamlList) => yamlList?.map((dynamic e) => '$e').toList();

  void generateEnv() {
    // loop through all the path for .env files
    for (final String path in envPaths) {
      // TODO(berbsd): indicate in comments if value is overriden
      for (final Flavor flavor in environments.values) {
        final StringBuffer buffer = StringBuffer()
          ..writeln(_header(comment: kCommentEnv, flavorName: flavor.name))
          ..writeln(flavor.keys.values.map((v) => v.toEnv).join('\n'));

        final Directory outputDir = Directory(join(Directory.current.path, path));
        if (!outputDir.existsSync()) {
          throw Exception('Directory does not exist - $outputDir');
        }

        final String filename = '.env.${flavor.name}';
        File(join(outputDir.path, filename)).writeAsStringSync(buffer.toString());
      }
    }
  }

  void generateDart() {
    // loop through all the path for .env files
    for (final String path in classPath) {
      // remove 'default' flavor from the list of environments
      final Flavor? defaultEnv = environments.remove('default');

      for (final Flavor flavor in environments.values) {
        final StringBuffer buffer = StringBuffer();

        final HashMap<String, EnvVar> envKeys = HashMap.from(flavor.keys);
        // add to the environments the values from 'defautl' that are not overwritten
        defaultEnv?.keys.forEach((key, variable) => envKeys.putIfAbsent(key, () => variable));

        buffer
          ..writeln(_header(comment: kCommentDart, flavorName: flavor.name))
          ..writeln('class $className  {\n')
          ..writeln(envKeys.values.map((v) => v.toDart).join('\n'))
          ..writeln('\n}\n');

        final classDefinition = DartFormatter().format(buffer.toString());

        // check that directory exist or fails.
        final Directory outputDir = Directory(join(Directory.current.path, path));
        if (!outputDir.existsSync()) {
          throw Exception('Directory does not exist - $outputDir');
        }

        final String filename = '${camelCaseToLowerUnderscore(className)}_${flavor.name.toLowerCase()}.dart';
        File(join(outputDir.path, filename)).writeAsStringSync(classDefinition);
      }
    }
  }
}

String _header({required String comment, required String flavorName}) {
  return '''

$comment AUTO-GENERATED - DO NOT MODIFY THIS FILE
$comment ----------------------------------------
$comment FLAVOR: $flavorName
''';
}
